/*
 * packetmodeller/netdevice.c
 *
 * Copyright (C) 2008-2009 Reto Gantenbein
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <linux/if_tun.h>
#include <netinet/in.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include "logger.h"
#include "netdevice.h"
#include "packet_modeller.h"
#include "typedef.h"

/** Fill interface request with information of given device */
int
get_ifr_by_dev(struct ifreq *ifr_ext, const ifname *dev_name)
{
    int fd, i;
    int ifrs_size = 10;
    struct ifconf ifc;
    struct ifreq *ifr_p;
    struct ifreq ifrs[ifrs_size];

    if ((fd = socket(PF_INET, SOCK_DGRAM, 0)) == -1)
    {
        print_dbg("%s:%i - Error opening socket (%s)\n", \
            __FILE__, __LINE__, strerror(errno));
        return -1;
    }
    memset(&ifrs, 0, sizeof(ifrs));
    ifc.ifc_len = sizeof(ifrs);
    ifc.ifc_req = ifrs;
    if (ioctl(fd, SIOCGIFCONF, (void *) &ifc) < 0)
    {
        close(fd);
        print_dbg("%s:%i - Couldn't fetch device configuration (%s)\n", \
            __FILE__, __LINE__, strerror(errno));
        return -1;
    }
    if (ifc.ifc_len == sizeof(ifrs))
    {
        close(fd);
        print_dbg("%s:%i - Couldn't fetch all device information \
            (ifc_len too small)\n", __FILE__, __LINE__);
        return -1;
    }
    ifrs_size = ifc.ifc_len / sizeof(struct ifreq);
    for (i = 0; i < ifrs_size; i++)
    {
        ifr_p = ((struct ifreq*) ifc.ifc_req) + i;
        if (strcmp((char *) dev_name, (*ifr_p).ifr_name) == 0)
        {
            memcpy(ifr_ext, ifr_p, sizeof(struct ifreq));
            return 0;
        }
    }
    return -1;
}

/**
 * Return the mac address of the given interface
 * @param mac - 6 bytes char array to fill in MAC address
 * @param dev_name - device of requested address
 */
int get_mac_by_dev(unsigned char *mac, const ifname *dev_name)
{
    int fd;
    struct ifreq ifr;

    if (get_ifr_by_dev(&ifr, dev_name) == 0)
    {
        if ((fd = socket(PF_INET, SOCK_DGRAM, 0)) == -1)
        {
            print_dbg("%s:%i - Error opening socket (%s)\n", \
                __FILE__, __LINE__, strerror(errno));
            return -1;
        }
        if (ioctl(fd, SIOCGIFHWADDR, &ifr) == 0)
        {
            memcpy(mac, ifr.ifr_hwaddr.sa_data, 6);
            return 0;
        } else {
            print_dbg("%s:%i - ", __FILE__, __LINE__);
            print_dbg("Couldn't get MAC address of interface ");
            print_dbg("%s (%s)\n", (char *) dev_name, strerror(errno));
            return -1;
        }
    } else {
        return -1;
    }
}

/**
 * Return ip address to a given interface
 * @param addr - struct to fill with ip address
 * @param dev_name - device of requested address
 */
int get_ip_by_dev(struct in_addr *addr, const ifname *dev_name)
{
    struct ifreq ifr;

    if (get_ifr_by_dev(&ifr, dev_name) == 0)
    {
        if (addr != NULL) {
            // if we only check for valid interface we don't need addr struct
            struct sockaddr_in *sockaddr =
                (struct sockaddr_in *) &(ifr.ifr_addr);
            memcpy(addr, (struct in_addr *) &(*sockaddr).sin_addr, \
                sizeof(struct in_addr));
        }
        return 0;
    } else {
        return -1;
    }
}

/**
 *  Initialize address struct with data from given interface and port
 *  @param addr - struct to fill
 *  @param dev - device to retrieve information from
 *  @param port - port to fill in addr (0 = random port)
 */
int
get_addr(struct sockaddr_in *addr, const ifname *dev, const uint16_t port)
{
    get_ip_by_dev(&(*addr).sin_addr, dev);
    addr->sin_family = AF_INET;
    addr->sin_port = htons(port);
    memset(addr->sin_zero, '\0', sizeof(addr->sin_zero));
    return 0;
}

/**
 * Create UDP socket and bind to given interface
 * @param dev - device name (e.g. eth0)
 * @param port - port to bind to
 * @return socket device file descriptor for network device
 */
int
alloc_net_socket(const ifname *dev, const uint16_t port)
{
    int fd = 0;
    struct sockaddr_in *local_addr;

    // use SOCK_DGRAM (=> UDP)
    if ((fd = socket(PF_INET, SOCK_DGRAM, 0)) == -1)
    {
        print_dbg("%s:%i - Error opening socket (%s)\n", \
            __FILE__, __LINE__, strerror(errno));
    }
    local_addr = calloc(1, sizeof(struct sockaddr_in));
    get_addr(local_addr, dev, port);
    if (bind(fd, (struct sockaddr *) local_addr, \
        (socklen_t) sizeof(struct sockaddr)) == -1)
    {
        print_dbg("%s:%i - Couldn't bind to local interface %s (%s)\n", \
            __FILE__, __LINE__, (char *) dev, strerror(errno));
    }
    free(local_addr);
    return fd;
}

/**
 * Return character device file descriptor for TAP device
 * @param dev - device name (e.g. v_ath0)
 */
int alloc_tap_socket(const ifname *dev)
{
    int fd = 0, err = 0;
    struct ifreq ifr;

    fd = open("/dev/net/tun", O_RDWR);
    if( fd < 0 )
    {
        print_dbg("Error opening /dev/net/tun (%s)\r\n", strerror(errno));
        return -1;
    }
    memset(&ifr, 0, sizeof(ifr));
    ifr.ifr_flags = IFF_TAP;    // set flag for TAP device
    ifr.ifr_flags = IFF_NO_PI;  // we don't need packet info (?)
    if (*dev)
    {
       memcpy(ifr.ifr_name, dev, IFNAMSIZ);
    }
    err = ioctl(fd, TUNSETIFF, (void *) &ifr);
    err = ioctl(fd, TUNSETNOCSUM, 1);
    if (err < 0)
    {
        close(fd);
        print_dbg("Error ioctl on %s (%s)\r\n", (char *) dev, strerror(errno));
        return -1;
    }
    strcpy((char *) dev, ifr.ifr_name);
    return fd;
}
