/*
 * wlanmodel/crawsocketrtscheduler.cc
 *
 * Copyright (C) 2008-2009 Reto Gantenbein
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <cerrno>
#include "crawsocketrtscheduler.h"

Register_Class(cRAWSocketRTScheduler);

/**
 * Define << operator for type timeval
 */
inline std::ostream& operator<<(std::ostream& ev, const timeval& tv)
{
    return ev << (unsigned long)tv.tv_sec << "s" << tv.tv_usec << "us";
}

/**
 * Constructor
 */
cRAWSocketRTScheduler::cRAWSocketRTScheduler() : cScheduler()
{
    listenerSocket = INVALID_SOCKET;
    connSocket = INVALID_SOCKET;
}

/**
 * Destructor
 */
cRAWSocketRTScheduler::~cRAWSocketRTScheduler()
{
}

/**
 * Called at the beginning of a simulation run.
 */
void
cRAWSocketRTScheduler::startRun()
{
    if (initsocketlibonce()!=0)
        throw new cRuntimeError("cRAWSocketRTScheduler: Cannot initialize \
socket library");

    gettimeofday(&baseTime, NULL);

    module = NULL;
    notificationMsg = NULL;
    recvBuffer = NULL;
    recvBufferSize = 0;
    numBytesPtr = NULL;

    port = ev.config()->getAsInt("General", "socketrtscheduler-port", 2424);
    setupListener();
}

/**
 * Called at the end of a simulation run.
 */
void
cRAWSocketRTScheduler::endRun()
{
}

/**
 * Setup listening socket for receiving node traffic
 */
void
cRAWSocketRTScheduler::setupListener()
{
    // open UDP socket
    listenerSocket = socket(AF_INET, SOCK_DGRAM, 0);
    if (listenerSocket == INVALID_SOCKET)
        throw new cRuntimeError("cRAWSocketRTScheduler: cannot create socket");

    sockaddr_in sinInterface;
    sinInterface.sin_family = AF_INET;
    sinInterface.sin_addr.s_addr = INADDR_ANY;
    sinInterface.sin_port = htons(port);
    if (bind(listenerSocket, (sockaddr*) &sinInterface, sizeof(sockaddr_in)) \
        == SOCKET_ERROR)
        throw new cRuntimeError("cRAWSocketRTScheduler: socket bind() failed");
}

/**
 * Called every time the user hits the Run button in Tkenv.
 * Use this callback to pin current simulation time to current wall clock time.
 */
void
cRAWSocketRTScheduler::executionResumed()
{
    gettimeofday(&baseTime, NULL);
    baseTime = timeval_substract(baseTime, sim->simTime());
}

void
cRAWSocketRTScheduler::setInterfaceModule(cModule *mod, cMessage *notifMsg, \
    char *buf, int bufSize, int *nBytesPtr)
{
    if (module)
        throw new cRuntimeError("cRAWSocketRTScheduler: setInterfaceModule() \
already called");
    if (!mod || !notifMsg || !buf || !bufSize || !nBytesPtr)
        throw new cRuntimeError("cRAWSocketRTScheduler: setInterfaceModule(): \
arguments must be non-NULL");

    module = mod;
    notificationMsg = notifMsg;
    recvBuffer = buf;
    recvBufferSize = bufSize;
    numBytesPtr = nBytesPtr;
    *numBytesPtr = 0;
}

/**
 * Listen for incoming network traffic until timeout exceeds
 */
bool cRAWSocketRTScheduler::receiveWithTimeout(long usec)
{
    // prepare sets for select()
    fd_set readFDs, writeFDs, exceptFDs;
    FD_ZERO(&readFDs);
    FD_ZERO(&writeFDs);
    FD_ZERO(&exceptFDs);

    FD_SET(listenerSocket, &readFDs);

    timeval timeout;
    timeout.tv_sec = 0;
    timeout.tv_usec = usec;

    if (select(FD_SETSIZE, &readFDs, &writeFDs, &exceptFDs, &timeout) > 0)
    {
        if (listenerSocket != INVALID_SOCKET && \
            FD_ISSET(listenerSocket, &readFDs))
        {
            // receive from listenerSocket
            sockaddr_in sinRemote;
            int addrSize = sizeof(sinRemote);
            char *bufPtr = recvBuffer + (*numBytesPtr);
            int bufLeft = recvBufferSize - (*numBytesPtr);
            if (bufLeft <= 0)
                throw new cRuntimeError("cRAWSocketRTScheduler: interface \
module's recvBuffer is full");
            int nBytes = recvfrom(listenerSocket, bufPtr, bufLeft, 0, \
                (sockaddr*) &sinRemote, (socklen_t*) &addrSize);
            if (nBytes == SOCKET_ERROR)
            {
                ev << "cRAWSocketRTScheduler: error receiving socket" << \
                    sock_errno() << "\n";
                connSocket = INVALID_SOCKET;
            } else {
                // schedule notificationMsg for the interface module
                ev << "cRAWSocketRTScheduler: " << nBytes \
                    << " bytes received\n";
                (*numBytesPtr) += nBytes;

                timeval curTime;
                gettimeofday(&curTime, NULL);
                notificationMsg->setName("rtmsg");
                notificationMsg->setTimestamp(curTime.tv_sec + \
                    curTime.tv_usec * 1e-6);
                curTime = timeval_substract(curTime, baseTime);
                simtime_t t = curTime.tv_sec + curTime.tv_usec * 1e-6;
        // TBD assert that it's somehow not smaller than previous event's time
                notificationMsg->setArrival(module,-1,t);
                simulation.msgQueue.insert(notificationMsg);
                return true;
            }
        }
    }
    return false;
}

int
cRAWSocketRTScheduler::receiveUntil(const timeval& targetTime)
{
    // if there's more than 200ms to wait, wait in 100ms chunks
    // in order to keep UI responsiveness by invoking ev.idle()
    timeval curTime;
    gettimeofday(&curTime, NULL);
    while (targetTime.tv_sec-curTime.tv_sec >=2 ||
           timeval_diff_usec(targetTime, curTime) >= 200000)
    {
        if (receiveWithTimeout(100000)) // 100ms
            return 1;
        if (ev.idle())
            return -1;
        gettimeofday(&curTime, NULL);
    }

    // difference is now at most 100ms, do it at once
    long usec = timeval_diff_usec(targetTime, curTime);
    if (usec>0)
        if (receiveWithTimeout(usec))
            return 1;
    return 0;
}

/**
 * The scheduler function -- it should return the next event to be processed.
 * Normally (with sequential execution) it just returns msgQueue.peekFirst().
 * With parallel and/or real-time simulation, it is also the scheduler's task
 * to synchronize with real time and/or with other partitions.
 */
cMessage *cRAWSocketRTScheduler::getNextEvent()
{
    ev << "Processing next event\n";

    // assert that we've been configured
    if (!module)
        throw new cRuntimeError("cRAWSocketRTScheduler: setInterfaceModule() \
not called: it must be called from a module's initialize() function");

    // calculate target time
    timeval targetTime;
    cMessage *msg = sim->msgQueue.peekFirst();
    if (!msg)
    {
        // if there are no events, wait until something comes from outside
        // TBD: obey simtimelimit, cpu-time-limit
        targetTime.tv_sec = LONG_MAX;
        targetTime.tv_usec = 0;
    }
    else
    {
        // use time of next event
        simtime_t eventSimtime = msg->arrivalTime();
        targetTime = timeval_add(baseTime, eventSimtime);
    }

    // if needed, wait until that time arrives
    timeval curTime;
    gettimeofday(&curTime, NULL);
    if (timeval_greater(targetTime, curTime))
    {
        int status = receiveUntil(targetTime);
        if (status == -1)
            return NULL; // interrupted by user
        if (status == 1)
            msg = sim->msgQueue.peekFirst(); // received something
    }
    else
    {
        // we're behind -- customized versions of this class may
        // alert if we're too much behind, whatever that means
    }

    // ok, return the message
    return msg;
}

int
cRAWSocketRTScheduler::sendMessage(const char *msg, size_t numBytes, \
    struct sockaddr_in *node_addr)
{
    int bytesSent;
    if (listenerSocket == INVALID_SOCKET)
        throw new cRuntimeError("cRAWSocketRTScheduler: sendMessage(): \
no connection");

    bytesSent = sendto(listenerSocket, msg, numBytes, 0, \
        (struct sockaddr *) node_addr, sizeof(struct sockaddr_in));
    ev << "cRAWSocketRTScheduler: " << bytesSent << " bytes sent (" << \
        strerror(errno) << ")\n";
    if (bytesSent <= 0)
    {
        return SOCKET_ERROR;
    } else {
        return 0;
    }
}
