/*
 * wlanmodel/logger.cc
 *
 * Copyright (C) 2008-2009 Reto Gantenbein
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <arpa/inet.h>
#include <net/ethernet.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "typedef.h"
#include "logger.h"
#include "message.h"
#include "packetproxy.h"

/** Print raw data in hex to stdout */
void
dump_bytes(const char *data, int len)
{
    int i;

    printf("(%d bytes) data: ", len);
    for(i = 1; i <= len; i++)
    {
        printf("%02hhx ", *(data + (i-1)));
        if (i % 16 == 0)
            printf("\n                 ");
    }
    printf("\n");
}

/** Print message header struct */
void
dump_header(struct msg_header *header)
{
    int i;

    printf("(%i bytes) type: ", sizeof(header->type));
    switch (header->type)
    {
        case NODE_MSG:
            printf("Node_Msg");
            break;
        case MODEL_MSG:
            printf("Model_Msg");
            break;
        case REGISTER_MSG:
            printf("Register_Msg");
            break;
        case ACK_MSG:
            printf("Acknowledge_Msg");
            break;
        default:
            printf("unknown message");
            break;
    }
    printf(" (%i)\n", header->type);
    printf("(%i bytes) id: %s\n", sizeof(header->id), header->id);
}

/** Print information from register message */
void
dump_register_info(struct register_msg *msg)
{
    printf("(%i bytes) msg_id: %i\n", sizeof(msg->msg_id), msg->msg_id);
    printf("(%i bytes) name: %s\n", sizeof(msg->name), msg->name);
    printf("(%i bytes) ip: %s\n", sizeof(msg->ip), inet_ntoa(msg->ip));
    printf("(%i bytes) port: %i\n", sizeof(msg->port), ntohs(msg->port));
    printf("(%i bytes) virt_mac: %s\n", sizeof(msg->virt_mac), msg->virt_mac);
}

/** Print information about wlan parameters sent to model */
void
dump_wlan_info(struct node_msg *msg)
{
    printf("(%i bytes) channel: %i\n", sizeof(msg->channel), msg->channel);
    printf("(%i bytes) transmit_power: %i\n", sizeof(msg->transmit_power), \
        msg->transmit_power);
    printf("(%i bytes) pkt_length: %i\n", sizeof(msg->pkt_length), \
        msg->pkt_length);
}

/** Print IP header and return packet length */
u_int16_t
dump_ethernet_header(char* ehdr)
{
    //structs and types defined in "net/ethernet.h"
    int i;
    struct ether_header *ether_hdr = (struct ether_header *) ehdr;

    printf("(%i bytes) header: ether_dhost ", sizeof(struct ether_header));
    for (i = 0; i < ETH_ALEN; i++)
    {
        printf("%02hhx", (*ether_hdr).ether_dhost[i]);
        if (i < (ETH_ALEN - 1))
            printf(":");
    }
    printf("\n                   ether_shost ");
    for (i = 0; i < ETH_ALEN; i++)
    {
        printf("%02hhx", (*ether_hdr).ether_shost[i]);
        if (i < (ETH_ALEN - 1))
            printf(":");
    }
    printf("\n                   ether_type ");
    u_int16_t ether_type = ntohs((*ether_hdr).ether_type);
    if (ether_type == ETHERTYPE_PUP)
        printf("Xerox PUP\n");
    if (ether_type == ETHERTYPE_IP)
        printf("IP\n");
    if (ether_type == ETHERTYPE_ARP)
        printf("ARP\n");
    if (ether_type == ETHERTYPE_REVARP)
        printf("Reverse ARP\n");

    return ether_type;
}

/** Print PacketModeller <-> WlanModel message */
void
dump_msg(const char *pkt, int frame_length)
{
    node_msg *node_pkt;
    model_msg *model_pkt;

    printf("****************************************\n");
    printf("dump_msg():\n");
    dump_header((msg_header *) pkt);
    switch(PacketProxy::getMessageType((msg_header *) pkt))
    {
        case NODE_MSG:
            node_pkt = (node_msg *) pkt;
            dump_wlan_info(node_pkt);
            dump_ethernet_header((char *) &node_pkt->wlan_pkt);
            printf("****************************************\n");
            printf("dump_bytes():\n");
            dump_bytes((char *) &node_pkt->wlan_pkt, frame_length);
            break;
        case MODEL_MSG:
            model_pkt = (model_msg *) pkt;
            dump_ethernet_header((char *) &model_pkt->wlan_pkt);
            printf("****************************************\n");
            printf("dump_bytes():\n");
            dump_bytes((char *) &model_pkt->wlan_pkt, frame_length);
            break;
        case REGISTER_MSG:
            dump_register_info((register_msg *) pkt);
            break;
    }

    printf("****************************************\n");
}

/** Do nothing but imitate the printf() signature */
int
ignore_dbg(const char* text, ...) {
    return 0;
}
